from logging.config import dictConfig

from flask import Flask
from flask import g
from flask_sqlalchemy import SQLAlchemy

from nldcsc.plugins.sql_migrate import Config, SqlMigrate as _SqlMigrate
from nldcsc.plugins.sql_migrate.config.constants import LOGGING_CONFIG

dictConfig(LOGGING_CONFIG)


class _MigrateConfig(object):
    def __init__(self, migrate: "SqlMigrate", db: SQLAlchemy, **kwargs):
        self.migrate = migrate
        self.db = db
        self.directory = migrate.directory
        self.configure_args = kwargs


class SqlMigrate(_SqlMigrate):
    def __init__(
        self,
        app: Flask = None,
        db: SQLAlchemy = None,
        directory: str = "migrations",
        command: str = "db",
        compare_type: bool = True,
        render_as_batch: bool = True,
        **kwargs,
    ):
        """


        Args:
            app:
            db:
            directory:
            command:
            compare_type:
            render_as_batch:
            **kwargs:
        """

        self.sql_alchemy = db

        super().__init__(
            directory=directory,
            command=command,
            compare_type=compare_type,
            render_as_batch=render_as_batch,
            **kwargs,
        )

        if app is not None and self.sql_alchemy is not None:
            self.init_app(app, self.sql_alchemy, directory)

    def init_app(
        self,
        app: Flask,
        db: SQLAlchemy = None,
        directory: str = None,
        command: str = None,
        compare_type: bool = None,
        render_as_batch: bool = None,
        **kwargs,
    ):
        """


        Args:
            app:
            db:
            directory:
            command:
            compare_type:
            render_as_batch:
            **kwargs:

        Returns:

        """

        self.sql_alchemy = db or self.sql_alchemy
        self.metadata = self.sql_alchemy.metadata

        with app.app_context():
            self.db = self.sql_alchemy.engine

        super().init_app(
            self.db, directory, command, compare_type, render_as_batch, **kwargs
        )

        if not hasattr(app, "extensions"):
            app.extensions = {}

        app.extensions["migrate"] = _MigrateConfig(
            self, self.sql_alchemy, **self.alembic_ctx_kwargs
        )

        from .cli import db as db_cli_group

        app.cli.add_command(db_cli_group, name=self.command)

    def get_config(self, directory=None, x_arg=None, opts=None) -> Config:
        """


        Args:
            directory:
            x_arg:
            opts:

        Returns:

        """

        if x_arg is not None:
            if not isinstance(x_arg, list):
                x_arg = list(x_arg)
            x_arg.extend(getattr(g, "x_arg", []))
        else:
            x_arg = getattr(g, "x_arg", [])

        return super().get_config(directory, x_arg, opts)



import asyncio
from collections import defaultdict
from collections.abc import Callable
from functools import partial, wraps
import math
import os
from typing import (
    IO,
    Any,
    DefaultDict,
    Iterable,
    TypeVar,
    ParamSpec,
    Awaitable,
)
from aiofiles.threadpool.text import AsyncTextIOWrapper

from httpx import Response
from nldcsc.httpx_apis.base_class.httpx_base_class import HttpxBaseClass

T = TypeVar("T")
P = ParamSpec("P")


def signature_of(_: Callable[P, Any]) -> Callable[[Callable[..., T]], Callable[P, T]]:
    def wrapper(wrapped: Callable[..., T]) -> Callable[P, T]:
        return wrapped

    return wrapper


def async_call(
    calls: Callable[P, Any], parser: Callable[[Response | str | Any], T] = None
) -> Callable[[Callable[..., Any]], Callable[P, Awaitable[Response | str | Any | T]]]:
    def wrapper(
        _: Callable[..., Any],
    ) -> Callable[P, Awaitable[Response | str | Any | T]]:
        @wraps(calls)
        async def inner(
            self: "DCSCTransferAPI", *args, **kwargs
        ) -> Awaitable[Response | str | Any]:
            if not self.use_async_client:
                raise AttributeError(
                    "You are attempting to call a async method whilst using a sync client!"
                )

            args, kwargs = calls(self, *args, **kwargs)

            r = await self.a_call(*args, **kwargs)

            return parser(r) if parser else r

        return inner

    return wrapper


def sync_call(
    calls: Callable[P, Any], parser: Callable[[Response | str | Any], T] = None
) -> Callable[[Callable[..., Any]], Callable[P, Response | str | Any | T]]:
    def wrapper(_: Callable[..., Any]) -> Callable[P, Response | str | Any | T]:
        @wraps(calls)
        def inner(self: "DCSCTransferAPI", *args, **kwargs) -> Response | str | Any | T:
            if self.use_async_client:
                raise AttributeError(
                    "You are attempting to call a sync method whilst using a async client!"
                )

            args, kwargs = calls(self, *args, **kwargs)

            r = self.call(*args, **kwargs)

            return parser(r) if parser else r

        return inner

    return wrapper


def json_parser(response: Response) -> dict[Any, Any]:
    return response.json()


class UploadHelper:
    def __init__(self, upload_info: dict):
        self.max_chunk_size: int = upload_info["max_chunk_size"]
        self.batch_id: str = None
        self.chunks: DefaultDict[str, list[str]] = None

    def prepare_batch(
        self,
        *,
        handles: Iterable[IO[bytes] | AsyncTextIOWrapper] = None,
        filenames: list[str] = None,
    ):
        if not filenames:
            filenames = []

        if handles:
            filenames.extend(os.path.basename(io.name) for io in handles)

        return filenames

    def set_batch(self, batch: dict):
        self.batch_id = batch["batch_id"]
        self.chunks = defaultdict(list)

        for chunk in batch["expected"]:
            self.chunks[chunk["file"]].append(chunk["chunk_id"])

    def get_from_batch(
        self, *, handle: IO[bytes] | AsyncTextIOWrapper = None, filename: str = None
    ):
        if not self.batch_id:
            raise KeyError

        if not any(self.chunks.values()):
            raise EOFError

        filename = os.path.basename(handle.name) if handle else filename

        return self.chunks[filename].pop(), filename

    def calculate_chunk_count(self, file_size: int):
        return math.ceil(file_size / self.max_chunk_size)


class DCSCTransferAPI(HttpxBaseClass):
    def __init__(
        self,
        baseurl,
        token,
        api_path="api",
        proxies=None,
        user_agent="HttpxBaseClass",
        use_async_client=True,
        **kwargs,
    ):
        super().__init__(
            baseurl, api_path, proxies, user_agent, use_async_client, **kwargs
        )

        self.myheaders = self.__default_headers
        self.set_header_field("Access-Token", token)

    @property
    def __default_headers(self) -> dict:
        """
        Property to return the default headers
        """
        # print("call")
        return {
            "Accept": "application/json",
            "User-Agent": f"{self.user_agent}",
        }

    def _get_file(self, uuid: str):
        """
        Gets a file by uuid
        """

        resource = f"files/{uuid}"

        return (self.methods.GET,), {"resources": resource}

    def _get_upload_info(self):
        resource = "uploads/info"

        return (self.methods.GET,), {"resources": resource}

    def _search_hash(self, hash: str, hash_type: str):
        resource = f"hashes/{hash_type}/{hash}"

        return (self.methods.GET,), {"resources": resource}

    def _search_hash_sign(self, hash: str): ...

    def _list_hashes(self, start: int = 0, size: int = 10):
        resource = "hashes/list"

        data = {"start": start, "size": size}

        return (self.methods.GET,), {"resources": resource, "params": data}

    def _iter_hashes(self, data: dict):
        def parse(data: dict):
            size = data.get("size", 10)
            start = data.get("start", 0)

            return data.get("files", []), start + size, size

        files, next, size = parse(data)

        while len(files) == size:
            data = yield files, next

            files, next, size = parse(data)

        yield files, None

    def _get_upload_info(self):
        resource = "uploads/info"

        return (self.methods.GET,), {"resources": resource}

    def _request_batch(
        self, files: list[str], comment: str = None, is_malware: bool = True
    ):
        resource = "uploads/parts"

        data = {"is_malware": is_malware, "files": files}

        if comment:
            data["comment"] = comment

        return (self.methods.POST,), {"resources": resource, "json": data}

    def _upload_chunk(
        self,
        batch_id: str,
        chunk_id: str,
        chunk_idx: int,
        chunk_count: int,
        filename: str,
        content: bytes,
    ):
        resource = "uploads/parts"

        data = {
            "batch_id": batch_id,
            "chunk_id": chunk_id,
            "chunk_idx": chunk_idx,
            "chunk_count": chunk_count,
        }

        files = {"content": (filename, content, "application/octet-stream")}

        return (self.methods.PUT,), {
            "resources": resource,
            "data": data,
            "files": files,
        }

    @signature_of(_get_file)
    @async_call(_get_file, json_parser)
    async def a_get_file(self): ...

    @signature_of(_get_file)
    @sync_call(_get_file, json_parser)
    def get_file(self): ...

    @signature_of(_search_hash_sign)
    @async_call(partial(_search_hash, hash_type="md5"), json_parser)
    async def a_search_md5(self): ...

    @signature_of(_search_hash_sign)
    @sync_call(partial(_search_hash, hash_type="md5"), json_parser)
    def search_md5(self): ...

    @signature_of(_search_hash_sign)
    @async_call(partial(_search_hash, hash_type="sha1"), json_parser)
    async def a_search_sha1(self): ...

    @signature_of(_search_hash_sign)
    @sync_call(partial(_search_hash, hash_type="sha1"), json_parser)
    def search_sha1(self): ...

    @signature_of(_search_hash_sign)
    @async_call(partial(_search_hash, hash_type="sha256"), json_parser)
    async def a_search_sha256(self): ...

    @signature_of(_search_hash_sign)
    @sync_call(partial(_search_hash, hash_type="sha256"), json_parser)
    def search_sha256(self): ...

    @signature_of(_list_hashes)
    @async_call(_list_hashes, json_parser)
    async def a_list_hashes(self): ...

    @signature_of(_list_hashes)
    @sync_call(_list_hashes, json_parser)
    def list_hashes(self): ...

    @signature_of(_get_upload_info)
    @async_call(_get_upload_info, json_parser)
    async def a_get_upload_info(self): ...

    @signature_of(_get_upload_info)
    @sync_call(_get_upload_info, json_parser)
    def get_upload_info(self): ...

    @signature_of(_request_batch)
    @async_call(_request_batch, json_parser)
    async def a_request_batch(self): ...

    @signature_of(_request_batch)
    @sync_call(_request_batch, json_parser)
    def request_batch(self): ...

    @signature_of(_upload_chunk)
    @async_call(_upload_chunk, json_parser)
    async def a_upload_chunk(self): ...

    @signature_of(_upload_chunk)
    @sync_call(_upload_chunk, json_parser)
    def upload_chunk(self): ...

    def iter_hashes(self, fetch_size: int = 10):
        generator = self._iter_hashes(self.list_hashes(0, fetch_size))

        files, start = next(generator)

        while True:
            yield from files

            if start is None:
                break

            files, start = generator.send(self.list_hashes(start, fetch_size))

    async def a_iter_hashes(self, fetch_size: int = 10):
        generator = self._iter_hashes(await self.a_list_hashes(0, fetch_size))

        files, start = next(generator)

        while True:
            for file in files:
                yield file

            if start is None:
                break

            files, start = generator.send(await self.a_list_hashes(start, fetch_size))

    async def a_upload_files(
        self,
        *file_handles: AsyncTextIOWrapper,
        comment: str = None,
        is_malware: bool = True,
    ):
        helper = UploadHelper(await self.a_get_upload_info())
        helper.set_batch(
            await self.a_request_batch(
                helper.prepare_batch(handles=file_handles), comment, is_malware
            )
        )

        async def upload(file_handle: AsyncTextIOWrapper):
            await file_handle.seek(0, os.SEEK_END)
            file_size = await file_handle.tell()
            await file_handle.seek(0, os.SEEK_SET)

            chunk_id, filename = helper.get_from_batch(handle=file_handle)
            chunk_count = helper.calculate_chunk_count(file_size)

            for idx in range(chunk_count):
                result = await self.a_upload_chunk(
                    helper.batch_id,
                    chunk_id,
                    idx,
                    chunk_count,
                    filename,
                    await file_handle.read(helper.max_chunk_size),
                )

            return {**result, "file": filename}

        return await asyncio.gather(
            *(upload(file_handle) for file_handle in file_handles)
        )

    def upload_files(
        self, *file_handles: IO[bytes], comment: str = None, is_malware: bool = True
    ):
        helper = UploadHelper(self.get_upload_info())
        helper.set_batch(
            self.request_batch(
                helper.prepare_batch(handles=file_handles), comment, is_malware
            )
        )

        results = []

        for file_handle in file_handles:
            file_handle.seek(0, os.SEEK_END)
            file_size = file_handle.tell()

            file_handle.seek(0, os.SEEK_SET)

            chunk_id, filename = helper.get_from_batch(handle=file_handle)
            chunk_count = helper.calculate_chunk_count(file_size)

            for idx in range(chunk_count):
                result = self.upload_chunk(
                    helper.batch_id,
                    chunk_id,
                    idx,
                    chunk_count,
                    filename,
                    file_handle.read(helper.max_chunk_size),
                )

            results.append({**result, "file": filename})
        return results



import argparse
import logging
import os
import re
import shutil
import sys
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from functools import wraps
from logging.config import dictConfig
from typing import Optional, Dict, List

from alembic import __version__ as __alembic_version__
from alembic import util, migration
from alembic.autogenerate import RevisionContext
from alembic.config import Config as AlembicConfig
from alembic.migration import RevisionStep
from alembic.operations import Operations, BatchOperations
from alembic.runtime.environment import ProcessRevisionDirectiveFn
from alembic.script.revision import _RevIdType
from alembic.util import CommandError
from dataclasses_json import config as json_config
from dataclasses_json import dataclass_json
from sqlalchemy import Engine, MetaData
from sqlalchemy import Row
from sqlalchemy.schema import ColumnCollectionConstraint

from nldcsc.generic.utils import exclude_optional_dict
from .config.constants import LOGGING_CONFIG, schema_migrations_table
from .fractured_alembic.runtime.environment import SqlEnvironmentContext
from .fractured_alembic.runtime.migration import FracturedMigrationContext
from .fractured_alembic.runtime.op_shell import (
    create_shell,
)
from .fractured_alembic.script.base import SqlScriptDirectory, SqlScriptDirectoryContext
from .utils.helpers import timestamp_to_strf_string

dictConfig(LOGGING_CONFIG)

logger = logging.getLogger(__name__)

alembic_version = tuple([int(v) for v in __alembic_version__.split(".")[0:3]])

__version__ = "0.1"

_IGNORE_TABLES = ["schema_migrations"]


@dataclass_json
@dataclass
class SchemaMigrationRow:
    id: Optional[int] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    version_num: Optional[int] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    name: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    duration: Optional[float] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    migrated: Optional[int] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def fill_from_row(self, row: Row):
        self.id = row.id
        self.version_num = row.version_num
        self.name = row.name
        self.duration = row.duration
        self.migrated = timestamp_to_strf_string(row.migrated)

    @property
    def get_as_list(self):
        return [self.id, self.version_num, self.name, self.duration, self.migrated]


def catch_errors(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        logger.debug(f"Running {f.__name__}")
        try:
            r = f(*args, **kwargs)
        except (CommandError, RuntimeError) as exc:
            logger.error("Error: " + str(exc))
            sys.exit(1)
        else:
            logger.debug(f"Success running {f.__name__}")
            return r

    return wrapped


class _MigrateConfig(object):
    def __init__(self, migrate: "SqlMigrate", db: Engine, **kwargs):
        self.migrate = migrate
        self.db = db
        self.directory = migrate.directory
        self.configure_args = kwargs


class Config(AlembicConfig):
    def __init__(self, *args, **kwargs):
        self.template_directory = kwargs.pop("template_directory", None)
        super().__init__(*args, **kwargs)

    def get_template_directory(self) -> str:
        """
        Request the path to the template directory.

        Returns:
            Path to the template directory.
        """
        if self.template_directory:
            return self.template_directory
        package_dir = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(package_dir, "templates")

    def __repr__(self):
        return "<< SqlMigrateAlembicConfig >>"


class SqlMigrate(object):
    def __init__(
        self,
        db: Engine = None,
        metadata: MetaData = None,
        directory: str = "migrations",
        command: str = "db",
        compare_type: bool = True,
        render_as_batch: bool = True,
        **kwargs,
    ):
        """


        Args:
            app:
            db:
            directory:
            command:
            compare_type:
            render_as_batch:
            **kwargs:
        """
        self.configure_callbacks = []
        self.db = db
        self.command = command
        self.metadata = metadata
        self.directory = str(directory)
        self.alembic_ctx_kwargs = kwargs
        self.alembic_ctx_kwargs["compare_type"] = compare_type
        self.alembic_ctx_kwargs["render_as_batch"] = render_as_batch

        self._dir_path = None
        self._current_config = None

        self._environment_context = None
        self._script_directory = None
        self._script_directory_context = None

        if db is not None:
            self.init_app(db, directory)

    @property
    def environment_context(self) -> SqlEnvironmentContext:
        """
        Get the `SqlEnvironmentContext` for the current app.

        Returns:
            `SqlEnvironmentContext`
        """
        if self._environment_context is None:
            self._environment_context = SqlEnvironmentContext(
                self.current_config, self.script_directory
            )

        return self._environment_context

    @property
    def script_directory(self) -> SqlScriptDirectory:
        """
        Get the `SqlScriptDirectory` for the current app.

        Returns:
            `SqlScriptDirectory`
        """
        if self._script_directory is None:
            self._script_directory = SqlScriptDirectory.from_config(
                self.current_config,
                additional_options={"context": self._script_directory_context},
            )

        return self._script_directory

    @property
    def current_config(self) -> Config:
        """
        Get the `Config` for the current app.

        Returns:
            `Config`
        """
        return self._current_config

    @property
    def migration_context(self) -> FracturedMigrationContext:
        """
        Get the `FracturedMigrationContext` for the current app.

        Returns:
            `FracturedMigrationContext`
        """
        return self.environment_context.get_context()

    @contextmanager
    def _get_env_context(
        self, ctx_opts: dict = None, ensure_migrations_table: bool = False
    ):
        with self.environment_context as env_context:
            connectable = self.db.engine

            if ctx_opts is None:
                ctx_opts = {}

            logger.debug(f"Received context opts {ctx_opts}")

            with connectable.connect() as connection:
                logger.debug(
                    "Configuring migration context from environment context..."
                )
                env_context.configure(
                    connection,
                    self.metadata,
                    **self.alembic_ctx_kwargs,
                    **ctx_opts,
                    include_object=self._include_this_object,
                    process_revision_directives=self._process_revision_directives,
                    target_metadata=self.metadata,
                    script_directory=self.script_directory,
                )

                if ensure_migrations_table:
                    self.migration_context._ensure_schema_migrations_table()

                with env_context.begin_transaction():
                    yield env_context

    @staticmethod
    def generate_rev_id():
        """
        Method to generate a revision id based on the current date and time.

        Returns:
            String with the current revision id.

        Examples:
            20240810131949
        """
        rev_id = datetime.now().strftime("%Y%m%d%H%M%S")
        return rev_id

    def _process_revision_directives(self, context, revision, directives):
        if getattr(self.current_config.cmd_opts, "autogenerate", False):
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []
                logger.warning("No changes in schema detected.")

    @staticmethod
    def _include_this_object(object, name, type_, reflected, compare_to) -> bool:
        """
        Method to exclude the custom tables from the migrations / migration scripts

        Args:
            object:
            name: Name of the object to exclude
            type_: Type of the object to exclude
            reflected:
            compare_to:

        Returns:
            True if excluded, False otherwise
        """
        if type_ == "table" and name in _IGNORE_TABLES:
            return False
        return True

    def execute_migration(self, ctx_opts: Optional[dict] = None) -> None:
        """
        Execute migration related actions.

        Args:
            ctx_opts:

        Returns:
            None
        """

        with self._get_env_context(ctx_opts, True) as env_context:
            logger.debug("Running run_migrations in migration context...")
            env_context.run_migrations()

    def init_app(
        self,
        db: Engine = None,
        directory: str = None,
        command: str = None,
        compare_type: bool = None,
        render_as_batch: bool = None,
        **kwargs,
    ):
        """


        Args:
            app:
            db:
            directory:
            command:
            compare_type:
            render_as_batch:
            **kwargs:

        Returns:

        """
        self.db = db or self.db
        self.command = command or self.command
        self.directory = str(directory or self.directory)
        self.alembic_ctx_kwargs.update(kwargs)
        if compare_type is not None:
            self.alembic_ctx_kwargs["compare_type"] = compare_type
        if render_as_batch is not None:
            self.alembic_ctx_kwargs["render_as_batch"] = render_as_batch

        self._script_directory_context = SqlScriptDirectoryContext(
            self.db, schema_migrations_table
        )

        try:
            self.get_config()
        except RuntimeError:
            pass

    def configure(self, f):
        """


        Args:
            f:

        Returns:

        """
        self.configure_callbacks.append(f)
        return f

    def call_configure_callbacks(self, config: Config) -> Config:
        """


        Args:
            config:

        Returns:

        """
        for f in self.configure_callbacks:
            config = f(config)
        return config

    def get_config(self, directory=None, x_arg=None, opts=None) -> Config:
        """


        Args:
            directory:
            x_arg:
            opts:

        Returns:

        """
        if directory is None:
            directory = self.directory
        directory = str(directory)
        config = Config(os.path.join(directory, "alembic.ini"))
        config.set_main_option("script_location", directory)
        if config.cmd_opts is None:
            config.cmd_opts = argparse.Namespace()
        for opt in opts or []:
            setattr(config.cmd_opts, opt, True)
        if not hasattr(config.cmd_opts, "x"):
            setattr(config.cmd_opts, "x", [])
            if x_arg is not None:
                if isinstance(x_arg, list) or isinstance(x_arg, tuple):
                    for x in x_arg:
                        config.cmd_opts.x.append(x)
                else:
                    config.cmd_opts.x.append(x_arg)

        self._current_config = self.call_configure_callbacks(config)

        logger.debug(f"Current config cmd opts: {self.current_config.cmd_opts}")

        return self.current_config

    @catch_errors
    def init(self, directory=None, multidb=False, template=None, package=False):
        """

        Args:
            directory:
            multidb:
            template:
            package:

        Returns:

        """
        logger.debug(f"Received parameters: {locals()}")

        if directory is None:
            directory = self.directory
        template_directory = None
        if template is not None and ("/" in template or "\\" in template):
            template_directory, template = os.path.split(template)
        config = Config(template_directory=template_directory)
        config.set_main_option("script_location", directory)
        config.config_file_name = os.path.join(directory, "alembic.ini")
        config = self._current_config = self.call_configure_callbacks(config)
        if multidb and template is None:
            template = "flask-multidb"
        elif template is None:
            template = "flask"

        if os.access(directory, os.F_OK) and os.listdir(directory):
            raise util.CommandError(
                "Directory %s already exists and is not empty" % directory
            )

        template_dir = os.path.join(config.get_template_directory(), template)
        if not os.access(template_dir, os.F_OK):
            raise util.CommandError("No such template %r" % template)

        if not os.access(directory, os.F_OK):
            logger.info(f"Creating directory {os.path.abspath(directory)!r}")
            os.makedirs(directory)

        versions = os.path.join(directory, "versions")
        logger.info(
            f"Creating directory {os.path.abspath(versions)!r}",
        )
        os.makedirs(versions)

        config_file: str | None = None
        for file_ in os.listdir(template_dir):
            file_path = os.path.join(template_dir, file_)
            if file_ == "alembic.ini.mako":
                assert config.config_file_name is not None
                config_file = os.path.abspath(config.config_file_name)
                if os.access(config_file, os.F_OK):
                    util.msg(
                        f"File {config_file!r} already exists, skipping",
                        **config.messaging_opts,
                    )
                else:
                    self.script_directory._generate_template(
                        file_path, config_file, script_location=directory
                    )
            elif os.path.isfile(file_path):
                output_file = os.path.join(directory, file_)
                self.script_directory._copy_file(file_path, output_file)

        if package:
            for path in [
                os.path.join(os.path.abspath(directory), "__init__.py"),
                os.path.join(os.path.abspath(versions), "__init__.py"),
            ]:
                logger.info(f"Adding {path!r}")
                with open(path, "w"):
                    pass

        assert config_file is not None
        logger.info(
            "Please edit configuration/connection/logging "
            f"settings in {config_file!r} before proceeding.",
        )

    @catch_errors
    def revision(
        self,
        config: Optional[Config] = None,
        message: Optional[str] = None,
        autogenerate: bool = False,
        sql: bool = False,
        head: str = "head",
        splice: bool = False,
        branch_label: Optional[_RevIdType] = None,
        version_path: Optional[str] = None,
        rev_id: Optional[str] = None,
        depends_on: Optional[str] = None,
        process_revision_directives: Optional[ProcessRevisionDirectiveFn] = None,
    ):
        """


        Args:
            config:
            message:
            autogenerate:
            sql:
            head:
            splice:
            branch_label:
            version_path:
            rev_id:
            depends_on:
            process_revision_directives:

        Returns:

        """
        logger.debug(f"Received parameters: {locals()}")

        if rev_id is None:
            rev_id = SqlMigrate.generate_rev_id()

        opts = ["autogenerate"] if autogenerate else None
        config = self.get_config(self.directory, opts=opts)

        command_args = dict(
            message=message,
            autogenerate=autogenerate,
            sql=sql,
            head=head,
            splice=splice,
            branch_label=branch_label,
            version_path=version_path,
            rev_id=rev_id,
            depends_on=depends_on,
        )
        revision_context = RevisionContext(
            config,
            self.script_directory,
            command_args,
            process_revision_directives=process_revision_directives,
        )

        environment = util.asbool(config.get_main_option("revision_environment"))

        if autogenerate:
            environment = True

            if sql:
                raise util.CommandError(
                    "Using --sql with --autogenerate does not make any sense"
                )

            def retrieve_migrations(rev, context):
                revision_context.run_autogenerate(rev, context)
                return []

        elif environment:

            def retrieve_migrations(rev, context):
                revision_context.run_no_autogenerate(rev, context)
                return []

        elif sql:
            raise util.CommandError(
                "Using --sql with the revision command when "
                "revision_environment is not configured does not make any sense"
            )

        if environment:
            self.execute_migration(
                ctx_opts={
                    "fn": retrieve_migrations,
                    "as_sql": sql,
                    "template_args": revision_context.template_args,
                    "revision_context": revision_context,
                }
            )

            # the revision_context now has MigrationScript structure(s) present.
            # these could theoretically be further processed / rewritten *here*,
            # in addition to the hooks present within each run_migrations() call,
            # or at the end of env.py run_migrations_online().

        scripts = [script for script in revision_context.generate_scripts()]

        # set the script_directory back to None -> the property will create a new script_directory when requested.
        self._script_directory = None

        if len(scripts) == 1:
            return scripts[0]
        else:
            return scripts

    @catch_errors
    def migrate(
        self,
        message: Optional[str] = None,
        autogenerate: bool = True,
        sql: bool = False,
        head: str = "head",
        splice: bool = False,
        branch_label: Optional[_RevIdType] = None,
        version_path: Optional[str] = None,
        rev_id: Optional[str] = None,
        depends_on: Optional[str] = None,
        process_revision_directives: Optional[ProcessRevisionDirectiveFn] = None,
        x_arg=None,
    ):
        """
        Alias for 'alembic revision --autogenerate'

        Args:
            message:
            autogenerate:
            sql:
            head:
            splice:
            branch_label:
            version_path:
            rev_id:
            depends_on:
            process_revision_directives:
            x_arg:

        Returns:

        """
        logger.debug(f"Received parameters: {locals()}")

        config = self.get_config(self.directory, opts=["autogenerate"], x_arg=x_arg)

        return self.revision(
            config=config,
            message=message,
            autogenerate=autogenerate,
            sql=sql,
            head=head,
            splice=splice,
            branch_label=branch_label,
            version_path=version_path,
            rev_id=rev_id,
            depends_on=depends_on,
            process_revision_directives=process_revision_directives,
        )

    @catch_errors
    def upgrade(
        self,
        revision: str = "head",
        sql: bool = False,
        tag: Optional[str] = None,
        x_arg=None,
        max_lookback_days: int = 30,
        sync: bool = False,
        fix_head: bool = False,
    ) -> None:
        """


        Args:
            revision:
            sql:
            tag:
            x_arg:
            max_lookback_days

        Returns:

        """
        logger.debug(f"Received parameters: {locals()}")

        self.get_config(
            self.directory, x_arg=x_arg
        )  # This makes sure that self.current_config has the latest values...

        if revision != "head" and fix_head:
            raise util.CommandError("Can't fix head if not upgrading to head!")

        starting_rev = None
        if ":" in revision:
            if not sql:
                raise util.CommandError("Range revision not allowed")
            starting_rev, revision = revision.split(":", 2)

        def upgrade(rev, context):
            revs = []

            if sync:
                revs.extend(self.script_directory._sync_revs(rev))

            new_revs = self.script_directory._upgrade_revs(
                revision, (), max_lookback_days
            )

            revs.extend(new_revs)

            return revs

        self.execute_migration(
            ctx_opts={
                "fn": upgrade,
                "as_sql": sql,
                "is_sync": sync,
                "starting_rev": starting_rev,
                "destination_rev": revision,
                "tag": tag,
                "fix_head": fix_head,
            }
        )

    @catch_errors
    def downgrade(
        self,
        revision: str = "-1",
        sql: bool = False,
        tag: Optional[str] = None,
        x_arg=None,
    ) -> None:
        """


        Args:
            revision:
            sql:
            tag:
            x_arg:

        Returns:

        """
        logger.debug(f"Received parameters: {locals()}")

        self.get_config(
            self.directory, x_arg=x_arg
        )  # This makes sure that self.current_config has the latest values...

        starting_rev = None
        if ":" in revision:
            if not sql:
                raise util.CommandError("Range revision not allowed")
            starting_rev, revision = revision.split(":", 2)

        def downgrade(rev, context):
            return self.script_directory._downgrade_revs(revision, rev)

        self.execute_migration(
            ctx_opts={
                "fn": downgrade,
                "as_sql": sql,
                "starting_rev": starting_rev,
                "destination_rev": revision,
                "tag": tag,
            }
        )

    @catch_errors
    def check(
        self, revision: str = None, max_lookback_days: int = 30, sync: bool = False
    ):
        logger.debug(f"Received parameters: {locals()}")

        self.get_config(
            self.directory
        )  # This makes sure that self.current_config has the latest values...

        @dataclass
        class Row:
            rev: RevisionStep

            @property
            def as_list(self):
                revision = self.rev.revision.revision
                name = self.rev.doc

                try:
                    date = datetime.strptime(revision, "%Y%m%d%H%M%S")
                except ValueError:
                    try:
                        date = datetime.strptime(
                            re.search(
                                r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}",
                                self.rev.revision.longdoc,
                            ).group(),
                            "%Y-%m-%d %H:%M:%S",
                        )
                    except (AttributeError, ValueError):
                        return [revision, name, None, None]

                age: timedelta = datetime.now() - date
                hours = age.seconds // 3600
                minutes = age.seconds // 60

                if age.days:
                    age = f"{age.days} day{'s' if age.days > 1 else ''}"
                elif hours:
                    age = f"{hours} hour{'s' if hours > 1 else ''}"
                elif minutes:
                    age = f"{minutes} minute{'s' if minutes > 1 else ''}"
                else:
                    age = f"{age.seconds} second{'s' if age.seconds > 1 else ''}"

                return [revision, name, date, age]

        @dataclass
        class IRow:
            instruction: object
            op: object
            args: tuple

            @property
            def as_list(self):
                operation = (
                    self.instruction.__name__.removesuffix("_class")
                    .replace("_", " ")
                    .capitalize()
                )

                first_arg = self.args[0]

                if isinstance(self.op, BatchOperations):
                    first_arg = f"{self.op.impl.table_name}:{first_arg}"

                extra_args = []

                if len(self.args) > 1:
                    for i in range(1, len(self.args)):
                        if isinstance(self.args[i], ColumnCollectionConstraint):
                            try:
                                extra_args.append(
                                    f"{type(self.args[i]).__name__}, {','.join(self.args[i].columns)}"
                                )
                                continue
                            except Exception:
                                pass
                        extra_args.append(str(self.args[i]))

                extra_args = "\n".join(extra_args)

                return [operation, first_arg, extra_args]

        if revision is not None:
            steps = [["Operation", "Argument", "Further arguments"]]

            def cb(cls, operation, *args, **kwargs):
                steps.append(IRow(cls, operation, args).as_list)

            create_shell(None, cb)

            with self._get_env_context() as env_context:
                with Operations.context(env_context._migration_context):
                    step = migration.MigrationStep.upgrade_from_script(
                        self.script_directory.revision_map,
                        self.script_directory.get_revision(str(revision)),
                    )

                    step.migration_fn()

            return steps

        else:
            data = [["rev", "name", "date", "age"]]

            with self._get_env_context() as env_context:
                if sync:
                    revs = self.script_directory._sync_revs(
                        env_context.get_context().get_current_heads()
                    )

                    revs.extend(
                        self.script_directory._upgrade_revs(
                            "head",
                            (),
                            max_lookback_days,
                        )
                    )
                else:
                    revs = self.script_directory._upgrade_revs(
                        "head",
                        env_context.get_context().get_current_heads(),
                        max_lookback_days,
                    )

                for rev in revs:
                    data.append(Row(rev).as_list)

            return data

    def drop(
        self,
    ) -> None:
        """
        Method to drop all tables from the database and remove the migration directory

        Returns:
            None
        """
        logger.debug(f"Received parameters: {locals()}")
        self.get_config(
            self.directory,
        )  # This makes sure that self.current_config has the latest values...

        logger.info("Dropping all tables and migration directory...")

        self.metadata.drop_all(self.db)

        with self._get_env_context() as env_context:
            logger.debug(
                f"Dropping migration tables; context configured: {self.migration_context}"
            )
            self.migration_context.drop_migrations_tables()

        try:
            shutil.rmtree(self.directory, True)
        except Exception:
            logger.error("Failed to drop migration directory.")

        logger.info("Done!!")

    def history(self, include_table_headers: bool = True) -> List[List[str | int]]:
        """
        Method to request the current migration history of the database

        Args:
            include_table_headers: Whether to include the table headers in the output

        Returns:
            List with the migration history of the database
        """
        logger.debug(f"Received parameters: {locals()}")

        self.get_config(
            self.directory,
        )  # This makes sure that self.current_config has the latest values...

        with self._get_env_context() as env_context:
            migration_data = self.migration_context.get_full_migrations_table_content()
            logger.debug(f"Got migration data: {migration_data}")

            if include_table_headers:
                table_data = [
                    [
                        x.name.title()
                        for x in self.migration_context._schema_migrations_table.columns._all_columns
                    ]
                ]
            else:
                table_data = []

            migration_rows = []
            for each in migration_data:
                data = SchemaMigrationRow()
                data.fill_from_row(each)
                migration_rows.append(data.get_as_list)
            table_data.extend(migration_rows)

        return table_data

    @staticmethod
    @catch_errors
    def set_logger_debug_format() -> None:
        """
        Method that sets the logger level to DEBUG and adjusts the log format to include the module names

        Intended to be used primarily from the CLI.

        Returns:
            None
        """
        formatter = logging.Formatter(
            "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
        )
        logger.setLevel(logging.DEBUG)

        logger.handlers[0].setFormatter(formatter)
        logger.handlers[0].setLevel(logging.DEBUG)
        logger.debug("Debug enabled")

    @staticmethod
    @catch_errors
    def get_version(print_stdout: bool = False) -> Dict[str, str]:
        """
        Print the current version of sql_migrate and alembic.

        Args:
            print_stdout: Print to stdout or return object

        Returns:
            Version dictionary containing the current version of sql_migrate and alembic.
        """
        config = Config()
        versions = f"SqlMigrate: {__version__}, Alembic: {__alembic_version__}"
        if print_stdout:
            config.print_stdout(versions)
        else:
            return {"SqlMigrate": __version__, "Alembic": __alembic_version__}

    @staticmethod
    @catch_errors
    def list_templates(print_stdout: bool = False) -> List[str]:
        """
        List available templates.

        Args:
            print_stdout: Print to stdout or return object

        Returns:
            List of available templates strings

        """

        config = Config()
        if print_stdout:
            config.print_stdout("Available templates:\n")
        template_names = []
        for tempname in sorted(os.listdir(config.get_template_directory())):
            template_names.append(tempname)
            with open(
                os.path.join(config.get_template_directory(), tempname, "README")
            ) as readme:
                synopsis = next(readme).strip()
            if print_stdout:
                config.print_stdout("%s - %s", tempname, synopsis)

        return template_names

    def __repr__(self):
        return f"<< {self.__class__.__name__ } >>"



import json
import logging
import os

from kafka import KafkaProducer, KafkaConsumer

from nldcsc.generic.utils import getenv_dict

# Effectively disabling the following loggers
logging.getLogger("kafka").setLevel(logging.CRITICAL)


class FlaskKafka(object):
    def __init__(
        self, app=None, producer: bool = True, consumer: bool = True, **kwargs
    ):
        self._producer = None
        self._consumer = None

        self.create_producer = producer
        self.create_consumer = consumer
        self.kwargs = kwargs

        self.kafka_url = os.getenv("KAFKA_URL", "localhost:9092")
        self.kafka_kwargs = getenv_dict("KAFKA_KWARGS", None)

        if app is not None:
            self.init_app(app)

    def init_app(self, app, **kwargs):
        self.kwargs.update(kwargs)

        if self.kafka_kwargs is not None:
            self.kwargs.update(self.kafka_kwargs)

        if self.create_producer:
            self._producer = KafkaProducer(
                bootstrap_servers=self.kafka_url,
                key_serializer=lambda k: k.encode("ascii"),
                value_serializer=lambda v: json.dumps(v).encode("ascii"),
                **self.kwargs,
            )

            if self.producer is not None:
                app.producer = self.producer

                if self.producer.bootstrap_connected():
                    app.logger.info(f"Producer connected to KAFKA!!")
                else:
                    app.logger.critical(
                        f"Connection to KAFKA failed!! Producer could not connect!"
                    )
            else:
                app.logger.error(f"Connection to KAFKA failed!!, no producer set up...")

        if self.create_consumer:
            self._consumer = KafkaConsumer(
                bootstrap_servers=self.kafka_url,
                key_deserializer=lambda k: k.decode("ascii"),
                value_deserializer=lambda v: json.loads(v.decode("ascii")),
                auto_offset_reset="earliest",
                **self.kwargs,
            )

            if self.consumer is not None:
                app.consumer = self.consumer

                if self.consumer.bootstrap_connected():
                    app.logger.info(f"Consumer connected to KAFKA!!")
                else:
                    app.logger.critical(
                        f"Connection to KAFKA failed!! Consumer could not connect!"
                    )
            else:
                app.logger.error(f"Connection to KAFKA failed!!, no consumer set up...")

    @property
    def producer(self):
        return self._producer

    @property
    def consumer(self):
        return self._consumer

    def __del__(self):
        if self.producer is not None:
            self.producer.close(5)

        if self.consumer is not None:
            self.consumer.close(5)

    def __repr__(self):
        return "<< FlaskKafka >>"



import os
import re
from collections import namedtuple
from datetime import date, datetime
from json import JSONDecodeError

import pytest

from nldcsc.generic.times import (
    timestampTOcalendarattrs,
    timestampTOdatestring,
    timestampTOdatetime,
    timestampTOdatetimestring,
    timestringTOdatetimestring,
    timestringTOtimestamp,
    datetimeTOtimestamp,
    dateTOtimestamp,
)
from nldcsc.generic.utils import (
    getenv_dict,
    getenv_bool,
    getenv_list,
    str2bool,
    generate_random_password,
    getenv_choice,
    getenv_str,
    reverse_from_named_tuple,
    _true_set,
    _false_set,
    __MANDATORY_VALUE__,
)


class TestGenericUtils:

    def test_get_env_str(self):
        ENV_VAL = "NLDCSC"
        ENV_KEY = "TEST_STR_ENV"

        os.environ[ENV_KEY] = ENV_VAL

        assert getenv_str(ENV_KEY) == ENV_VAL

    def test_get_env_str_missing(self):
        ENV_KEY = "TEST_STR_ENV"

        if ENV_KEY in os.environ:
            del os.environ[ENV_KEY]

        assert getenv_str(ENV_KEY) is None
        assert getenv_str(ENV_KEY, "DEFAULT") == "DEFAULT"

    def test_get_env_str_mandatory(self):
        ENV_VAL = "NLDCSC"
        ENV_KEY = "TEST_STR_ENV"

        os.environ[ENV_KEY] = ENV_VAL

        assert getenv_str(ENV_KEY, mandatory=True) == ENV_VAL

    def test_get_env_str_mandatory_missing(self):
        ENV_KEY = "TEST_STR_ENV"

        os.environ[ENV_KEY] = __MANDATORY_VALUE__

        with pytest.raises(ValueError, match=f"{ENV_KEY} is not set!"):
            getenv_str(ENV_KEY, mandatory=True)

    def test_get_env_choice(self):
        ENV_VAL = "YES"
        ENV_KEY = "TEST_STR_ENV"
        CHOICES = [ENV_VAL, "NO"]

        os.environ[ENV_KEY] = ENV_VAL

        assert getenv_choice(ENV_KEY, CHOICES) == ENV_VAL

    def test_get_env_choice_invalid(self):
        ENV_VAL = "1234"
        ENV_KEY = "TEST_STR_ENV"
        CHOICES = ["YES", "NO"]

        os.environ[ENV_KEY] = ENV_VAL

        with pytest.raises(
            ValueError,
            match=f"The variable {ENV_KEY} is not set to a valid choice, choices are {re.escape(str(CHOICES))}!",
        ):
            getenv_choice(ENV_KEY, CHOICES)

    def test_get_env_choice_mandatory(self):
        ENV_VAL = "YES"
        ENV_KEY = "TEST_STR_ENV"
        CHOICES = [ENV_VAL, "NO"]

        os.environ[ENV_KEY] = ENV_VAL

        assert getenv_choice(ENV_KEY, CHOICES, mandatory=True) == ENV_VAL

    def test_get_env_choice_mandatory_missing(self):
        ENV_KEY = "TEST_STR_ENV"
        CHOICES = ["YES", "NO"]

        os.environ[ENV_KEY] = __MANDATORY_VALUE__

        with pytest.raises(ValueError, match=f"{ENV_KEY} is not set!"):
            getenv_choice(ENV_KEY, CHOICES, mandatory=True)

    def test_get_env_choice_mandatory_invalid(self):
        ENV_KEY = "TEST_STR_ENV"
        CHOICES = [__MANDATORY_VALUE__, "NO"]

        os.environ[ENV_KEY] = __MANDATORY_VALUE__

        with pytest.raises(ValueError, match=f"{ENV_KEY} is not set!"):
            getenv_choice(ENV_KEY, CHOICES, mandatory=True)

    def test_get_env_dict(self):
        os.environ["DICT_ENV"] = '{"key1": "value1", "key2": "value2"}'
        os.environ["DICT_ENV_ERROR"] = "{'key1': 'value1', 'key2': 'value2'}"

        env_dict = getenv_dict("DICT_ENV", {"key": "default"})

        assert isinstance(env_dict, dict)
        assert env_dict["key1"] == "value1"

        env_dict = getenv_dict("DICT_ENV_NON_EXISTENT", {"key": "default"})

        assert isinstance(env_dict, dict)
        assert env_dict["key"] == "default"

        env_dict = getenv_dict("DICT_ENV_NON_EXISTENT")

        assert isinstance(env_dict, dict)
        assert len(env_dict) == 0

        with pytest.raises(JSONDecodeError):
            getenv_dict("DICT_ENV_ERROR", {"key": "default"})

    def test_get_env_list(self):
        os.environ["LIST_ENV"] = '["value1", "value2"]'
        os.environ["LIST_ENV_ERROR"] = "['value1', 'value2']"

        env_list = getenv_list("LIST_ENV", ["default"])

        assert isinstance(env_list, list)
        assert env_list[0] == "value1"

        env_list = getenv_list("LIST_ENV_NON_EXISTENT", ["default"])

        assert isinstance(env_list, list)
        assert env_list[0] == "default"

        env_list = getenv_list("LIST_ENV_NON_EXISTENT")

        assert isinstance(env_list, list)
        assert len(env_list) == 0

        with pytest.raises(JSONDecodeError):
            getenv_list("LIST_ENV_ERROR", ["default"])

    def test_get_env_bool(self):
        os.environ["BOOL_ENV_TRUE"] = "True"
        os.environ["BOOL_ENV_FALSE"] = "False"
        os.environ["BOOL_ENV_ERROR"] = "Blerf"
        os.environ["BOOL_ENV_DICT"] = '{"key1": "value1", "key2": "value2"}'

        env_bool = getenv_bool("BOOL_ENV_TRUE", "False")
        env_bool_false = getenv_bool("BOOL_ENV_FALSE", "False")

        assert isinstance(env_bool, bool)
        assert env_bool is True
        assert env_bool_false is False

        env_bool = getenv_bool("BOOL_ENV_NON_EXISTENT", "False")

        assert isinstance(env_bool, bool)
        assert env_bool is False

        with pytest.raises(ValueError):
            getenv_bool("BOOL_ENV_ERROR", "False")

        with pytest.raises(ValueError):
            getenv_bool("BOOL_ENV_DICT", "False")

    def test_str_to_bool(self):
        choices = {"yes", "true", "1"}

        assert str2bool(choices.pop()) is True
        assert str2bool("false") is False

    def test_str_to_bool_invalid(self):
        VALS = ("1234", 1, True, 1.3, [], (), {})

        for v in VALS:
            with pytest.raises(
                ValueError,
                match=f'Expected "{", ".join(_true_set | _false_set)}" -> received "{re.escape(str(v))}"',
            ):
                str2bool(v)

    def test_str_to_bool_invalid_no_raise(self):
        VALS = ("1234", 1, True, 1.3, [], (), {})

        for v in VALS:
            assert str2bool(v, False) is None

    def test_random_password_generator(self):
        assert len(generate_random_password()) == 32

    def test_reverse_from_named_tuple(self):
        test_tuple = namedtuple("test_tuple", ("A", "B", "C", "D"))(0, 1, "2", "3")

        assert reverse_from_named_tuple(test_tuple, 0) == "A"
        assert reverse_from_named_tuple(test_tuple, 1) == "B"
        assert reverse_from_named_tuple(test_tuple, "2") == "C"
        assert reverse_from_named_tuple(test_tuple, "3") == "D"

    def test_reverse_from_named_tuple_missing(self):
        test_tuple = namedtuple("test_tuple", ("A", "B"))(0, 1)

        with pytest.raises(
            KeyError,
            match=f"The requested index does not exist! Choices are {re.escape(str({0: 'A', 1: 'B'}))}",
        ):
            reverse_from_named_tuple(test_tuple, 3)


class TestGenericTimes:
    def test_timestringTOtimestamp(self):
        times = (
            ("12-05-2025", 1747008000),
            ("12-05-2025 14:45", 1747061100),
            ("2025-05-12 14:45:30", 1747061130),
            ("14:45:30 12-05-2025", 1747061130),
            ("2025-05-12", 1747008000),
            ("2025-05-12 14:45:30 UTC", 1747061130),
            ("2025-05-12T14:45:30", 1747061130),
            ("2025-05-12T14:45:30Z", 1747061130),
            ("2025-05-12T14:45:30+0000", 1747061130),
            ("2025-05-12T14:45:30.123456", 1747061130),
            ("2025-05-12T14:45:30,123456", 1747061130),
            ("2025-05-12T14:45:30.123456Z", 1747061130),
            ("2025-05-12T14:45:30,123456Z", 1747061130),
            ("2025-05-12T14:45:30.123456Z+0000", 1747061130),
            ("2025-05-12T14:45:30,123456Z+0000", 1747061130),
            ("2025-05-12 14:45:30.123456+0000", 1747061130),
        )

        for time, value in times:
            assert timestringTOtimestamp(time) == value

    def test_timestringTOtimestamp_invalid(self):
        time = "INVALID"

        assert timestringTOtimestamp(time) is False

    def test_datetimeToTimestamp(self):
        time = datetime(
            year=2025, month=5, day=12, hour=14, minute=45, second=30, microsecond=0
        )

        assert datetimeTOtimestamp(time) == 1747061130

    def test_dateTOtimestamp(self):
        time = date(year=2025, month=5, day=12)

        assert dateTOtimestamp(time) == 1747008000

    def test_timestampTOdatetime(self):
        time = 1747061130

        assert timestampTOdatetime(time) == datetime(
            year=2025, month=5, day=12, hour=14, minute=45, second=30, microsecond=0
        )

    def test_timestampTOdatetimestring(self):
        time = 1747061130

        assert timestampTOdatetimestring(time, vis=True) == "2025-05-12 14:45:30"
        assert timestampTOdatetimestring(time) == "2025-05-12T14:45:30Z"

    def test_timestampTOcalendarattrs(self):
        time = 1747061130

        assert timestampTOcalendarattrs(time) == {
            "day": "12",
            "month": "May",
            "weekday": "Monday",
        }

    def test_timestampTOdatestring(self):
        time = 1747061130

        assert timestampTOdatestring(time) == "12-05-2025"

    def test_timestringTOdatetimestring(self):
        time = "2025-05-12 14:45:30.123456+0000"

        assert timestringTOdatetimestring(time) == "2025-05-12T14:45:30Z"

    def test_timestringTOdatetimestring_invalid(self):
        time = "INVALID"

        assert timestringTOdatetimestring(time) is False



from typing import Any

from sqlalchemy.orm import mapped_column
from typing_extensions import Annotated

# this file should contain all common annotations used:
# - to keep things consistent the naming convention should be <type>_<identifier> -> str_64, str_128, int_pk, etc.

# ints
int_pk = Annotated[int, mapped_column(primary_key=True)]
big_int_pk = Annotated[int, mapped_column(primary_key=True)]
big_int = Annotated[int, "big_int"]

# strings
str_16 = Annotated[str, 16]
str_30 = Annotated[str, 30]
str_32 = Annotated[str, 32]
str_50 = Annotated[str, 50]
str_64 = Annotated[str, 64]
str_100 = Annotated[str, 100]
str_128 = Annotated[str, 128]
str_256 = Annotated[str, 256]
str_512 = Annotated[str, 512]
str_text = str

# dicts
dict_json = dict[str, Any]

# lists
list_json = list[Any]

# custom
uuid = Annotated[str, "uuid"]
uuid_pk = Annotated[str, mapped_column(primary_key=True), "uuid"]



from dataclasses import dataclass, field
from typing import Any, Optional
from dataclasses_json import DataClassJsonMixin, config


@dataclass
class FBFErrors(DataClassJsonMixin):
    errors: list[str] | str


@dataclass
class FBFInfo(DataClassJsonMixin):
    application_name: str = field(metadata=config(field_name="ApplicationName"))
    application_version: str = field(metadata=config(field_name="ApplicationVersion"))


@dataclass
class NCSCFeedInfo(DataClassJsonMixin):
    feeds: list[str]


@dataclass
class NCSCFeedUpdate(DataClassJsonMixin):
    name: str
    count: int


@dataclass
class NCSCFeedUpdateSquence(DataClassJsonMixin):
    timestamp: int
    updates: list[NCSCFeedUpdate]


@dataclass
class NCSCFeedUpdateSequenceList(DataClassJsonMixin):
    entries: list[NCSCFeedUpdateSquence]


@dataclass
class NCSCFeedIndex(DataClassJsonMixin):
    name: str
    request_id: str
    index: str
    total_count: int
    start_time: Optional[int] = None
    stop_time: Optional[int] = None


@dataclass
class NCSCFeedItems(DataClassJsonMixin):
    name: str
    entries: list[Any]
    entries_count: int
    total_count: int
    request_id: str
    limit: int
    offset: int
    last: bool


@dataclass
class NCSCFeedUpdateTS(DataClassJsonMixin):
    name: str
    timestamp: int



import contextlib
import logging
import warnings
from functools import wraps

import requests
from urllib3.exceptions import InsecureRequestWarning

from nldcsc.generic.utils import getenv_bool

old_merge_environment_settings = requests.Session.merge_environment_settings


# noinspection TryExceptPass
@contextlib.contextmanager
def do_ssl_verification(cert_check: bool = getenv_bool("SSO_TLS_VERIFICATION", "True")):
    opened_adapters = set()

    def merge_environment_settings(self, url, proxies, stream, verify, cert):
        # Verification happens only once per connection so we need to close
        # all the opened adapters once we're done. Otherwise, the effects of
        # verify=False persist beyond the end of this context manager.
        opened_adapters.add(self.get_adapter(url))

        settings = old_merge_environment_settings(
            self, url, proxies, stream, verify, cert
        )
        settings["verify"] = cert_check

        return settings

    requests.Session.merge_environment_settings = merge_environment_settings

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", InsecureRequestWarning)
            yield
    finally:
        requests.Session.merge_environment_settings = old_merge_environment_settings

        for adapter in opened_adapters:
            try:
                adapter.close()
            except Exception:
                pass


def ssl_verification(func=None):

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):

            logger = logging.getLogger(__name__)

            with do_ssl_verification(
                cert_check=getenv_bool("SSO_TLS_VERIFICATION", "True")
            ):
                logger.info(
                    f"Monkey patching SSL verification; "
                    f"setting verify: {getenv_bool('SSO_TLS_VERIFICATION', 'True')}"
                )

                return_value = func(*args, **kwargs)
                return return_value

        return inner

    if callable(func):
        return wrapper(func=func)
    return wrapper



from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import dataclass_json

from nldcsc.generic.times import timestringTOtimestamp


@dataclass_json
@dataclass
class OrganisationDetails:
    name: Optional[str] = None
    organization: Optional[str] = None
    street1: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    postalCode: Optional[str] = None
    country: Optional[str] = None
    countryCode: Optional[str] = None
    email: Optional[str] = None
    telephone: Optional[str] = None
    rawText: Optional[str] = None


@dataclass_json
@dataclass
class Registrant(OrganisationDetails): ...


@dataclass_json
@dataclass
class AdministrativeContact(OrganisationDetails): ...


@dataclass_json
@dataclass
class TechnicalContact(OrganisationDetails): ...


@dataclass_json
@dataclass
class NameServers:
    rawText: Optional[str] = None
    hostNames: Optional[List[str]] = None
    ips: Optional[List[str]] = None


@dataclass_json
@dataclass
class MetaData:
    status: Optional[str] = None
    rawText: Optional[str] = None
    parseCode: Optional[int] = None
    dataError: Optional[str] = None
    header: Optional[str] = None
    strippedText: Optional[str] = None
    footer: Optional[str] = None
    audit: Optional[dict[str, str]] = None
    customField1Name: Optional[str] = None
    customField1Value: Optional[str] = None
    registrarName: Optional[str] = None
    registrarIANAID: Optional[str] = None
    createdDateNormalized: Optional[str] = None
    updatedDateNormalized: Optional[str] = None
    expiresDateNormalized: Optional[str] = None
    customField2Name: Optional[str] = None
    customField3Name: Optional[str] = None
    customField2Value: Optional[str] = None
    customField3Value: Optional[str] = None


@dataclass_json
@dataclass
class RegistryData(MetaData):
    domainName: Optional[str] = None
    nameServers: Optional["NameServers"] = None
    whoisServer: Optional[str] = None
    createdDate: Optional[str] = None
    updatedDate: Optional[str] = None
    expiresDate: Optional[str] = None

    def __post_init__(self):
        if self.__getattribute__("nameServers") is not None:
            self.__setattr__(
                "nameServers", NameServers(**self.__getattribute__("nameServers"))
            )


@dataclass_json
@dataclass
class WhoisRecord(MetaData):
    createdDate: Optional[str] = None
    updatedDate: Optional[str] = None
    expiresDate: Optional[str] = None
    registrant: Optional["Registrant"] = None
    administrativeContact: Optional["AdministrativeContact"] = None
    technicalContact: Optional["TechnicalContact"] = None
    domainName: Optional[str] = None
    nameServers: Optional["NameServers"] = None
    registryData: Optional["RegistryData"] = None
    domainAvailability: Optional[str] = None
    contactEmail: Optional[str] = None
    domainNameExt: Optional[str] = None
    estimatedDomainAge: Optional[int] = None
    ips: Optional[List[str]] = None

    def __post_init__(self):
        if self.__getattribute__("registrant") is not None:
            self.__setattr__(
                "registrant", Registrant(**self.__getattribute__("registrant"))
            )
        if self.__getattribute__("administrativeContact") is not None:
            self.__setattr__(
                "administrativeContact",
                AdministrativeContact(**self.__getattribute__("administrativeContact")),
            )
        if self.__getattribute__("technicalContact") is not None:
            self.__setattr__(
                "technicalContact",
                TechnicalContact(**self.__getattribute__("technicalContact")),
            )
        if self.__getattribute__("nameServers") is not None:
            self.__setattr__(
                "nameServers", NameServers(**self.__getattribute__("nameServers"))
            )
        if self.__getattribute__("registryData") is not None:
            self.__setattr__(
                "registryData", RegistryData(**self.__getattribute__("registryData"))
            )

    @property
    def domain_is_available(self):
        if (
            self.domainAvailability is not None
            and self.domainAvailability == "UNAVAILABLE"
        ):
            return False
        else:
            return True

    @property
    def get_registrar(self) -> str | None:
        return self.registrarName

    @property
    def get_created_date(self) -> int:
        if self.registryData.createdDate:
            return timestringTOtimestamp(self.registryData.createdDate)
        else:
            return 0

    @property
    def get_updated_date(self) -> int:
        if self.registryData.updatedDate:
            return timestringTOtimestamp(self.registryData.updatedDate)
        else:
            return 0

    @property
    def get_expires_date(self) -> int:
        if self.registryData.expiresDate:
            return timestringTOtimestamp(self.registryData.expiresDate)
        else:
            return 0


@dataclass_json
@dataclass
class WhoisReverseNSRecord:
    name: str = ""
    first_seen: int = 0
    last_visit: int = 0


@dataclass_json
@dataclass
class WhoisReverseMXRecord:
    name: str = ""
    first_seen: int = 0
    last_visit: int = 0


@dataclass_json
@dataclass
class WhoisReverseIPRecord:
    name: str = ""
    first_seen: int = 0
    last_visit: int = 0



import os.path
from shutil import rmtree

import pytest
from sqlalchemy import inspect

from nldcsc.flask_plugins.flask_sql_migrate import SqlMigrate


@pytest.fixture()
def app():
    from tests.applications.sql_migrate_app import app

    app.config["TESTING"] = True

    yield app


@pytest.fixture()
def test_dir():
    dirs = [
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "..",
            "migrations",
        )
    ]

    # cleanup before incase some other test does not properly clean the test dirs.
    for dir in dirs:
        if os.path.exists(dir):
            rmtree(dir)

    yield os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")

    for dir in dirs:
        if os.path.exists(dir):
            rmtree(dir)


@pytest.fixture()
def db_handle(app):
    from tests.applications.sql_migrate_app import db

    with app.app_context():
        yield db

        # The test app uses an in memory database. If we close the connection it gets removed.
        db.engine.dispose()


@pytest.fixture
def sql_migrate(app):
    from tests.applications.sql_migrate_app import db

    migrate = SqlMigrate()
    migrate.init_app(app, db)

    with app.app_context():
        yield migrate


class TestSqlMigrate:
    def test_happy_migration_run(self, sql_migrate, test_dir):
        assert not os.path.exists(
            os.path.join(test_dir, sql_migrate.directory)
        ), "Migration directory should not exist!"

        sql_migrate.init()

        assert os.path.exists(
            os.path.join(test_dir, sql_migrate.directory)
        ), "Migration directory should exist!"

        assert (
            len(os.listdir(os.path.join(test_dir, sql_migrate.directory, "versions")))
            == 0
        ), "version dir should be empty!"

        sql_migrate.migrate()

        assert (
            len(
                [
                    f
                    for f in os.listdir(
                        os.path.join(test_dir, sql_migrate.directory, "versions")
                    )
                    if os.path.isfile(
                        os.path.join(test_dir, sql_migrate.directory, "versions", f)
                    )
                ]
            )
            == 1
        ), "1 migration should be present in the migration directory!"

    def test_happy_upgrade_run(self, sql_migrate, app, db_handle, test_dir):
        # this test uses the init and migrate, if that test fails, this one will as well.
        sql_migrate.init()
        sql_migrate.migrate()

        assert not inspect(db_handle.engine).has_table(
            "user"
        ), "user table should not exist!"

        sql_migrate.upgrade()

        assert inspect(db_handle.engine).has_table("user"), "user table should exist!"

    def test_happy_downgrade_run(self, sql_migrate, app, db_handle, test_dir):
        # this test uses the init and migrate, if that test fails, this one will as well.
        sql_migrate.init()
        sql_migrate.migrate()
        sql_migrate.upgrade()

        assert inspect(db_handle.engine).has_table("user"), "user table should exist!"

        sql_migrate.downgrade()

        assert not inspect(db_handle.engine).has_table(
            "user"
        ), "user table should not exist!"

    def test_happy_drop_run(self, sql_migrate, app, db_handle, test_dir):
        sql_migrate.init()
        sql_migrate.migrate()
        sql_migrate.upgrade()

        assert inspect(db_handle.engine).has_table("user"), "user table should exist!"
        assert os.path.exists(
            os.path.join(test_dir, sql_migrate.directory)
        ), "Migration directory should exist!"

        sql_migrate.drop()

        assert not inspect(db_handle.engine).has_table(
            "user"
        ), "user table should not exist!"
        assert not os.path.exists(
            os.path.join(test_dir, sql_migrate.directory)
        ), "Migration directory should not exist!"



import collections
from ipaddress import ip_address
from typing import List

from httpx import Response

from nldcsc.httpx_apis.base_class.httpx_base_class import HttpxBaseClass

q_types = collections.namedtuple(
    "q_types", ["A", "NS", "CNAME", "SOA", "PTR", "MX", "TXT", "AAAA"]
)(1, 2, 5, 6, 12, 15, 16, 28)


class DOHRequester(HttpxBaseClass):
    def __init__(
        self,
        baseurl: str,
        api_path: str = None,
        proxies: dict = None,
        user_agent: str = "Certex",
        **kwargs,
    ):
        super().__init__(
            baseurl=baseurl,
            api_path=api_path,
            proxies=proxies,
            user_agent=user_agent,
            **kwargs,
        )

        self.clear_headers()
        self.set_header_field("accept", "application/dns-json")

    @staticmethod
    def _convert_to_reverse_format(ipaddress: str) -> str:
        """
        Convert an IP address to its reverse DNS format.

        Parameters:
            ipaddress: The IP address to convert.

        Returns:
            str: The IP address in reverse DNS format.
        """
        return ip_address(ipaddress).reverse_pointer

    async def get_domain_list_by_type(
        self, domains: List[str], dns_type: int | str = 1  # A record
    ) -> List[Response]:
        """
        Get the list of domains with specified dns type.

        Args:
            domains: list of domain names
            dns_type: https://www.iana.org/assignments/dns-parameters/dns-parameters.xhtml#dns-parameters-4 or as
                      string representation: 'A', 'NS', 'CNAME', 'SOA', 'PTR', 'MX', 'TXT', 'AAAA'

        Returns:
            List of `Response` objects
        """
        if isinstance(dns_type, str):
            try:
                dns_type = getattr(q_types, dns_type)
            except AttributeError:
                self.logger.warning(
                    f"Requesting a dns type ({dns_type}) which is not in the q_types tuple; passing requested int"
                )
                pass

        if dns_type == q_types.PTR:
            domains = [self._convert_to_reverse_format(x) for x in domains]

        resources = [f"dns-query?name={each}&type={dns_type}" for each in domains]
        data = await self.a_call(self.methods.GET, resources=resources)
        return data

    async def get_bulk_domains(
        self, workload: dict[int | str, List[str]]
    ) -> List[Response]:
        """
        Get the list of domains with specified workload.

        Args:
            workload: dictionary with qtype as key (string or int representation, or '*' to query for all q_type) and
                      a list of domain names as value.

        Returns:
            List of `Response` objects
        """
        resources = []
        query_all_types = False

        for qtype, domain_list in workload.items():
            if isinstance(qtype, str):
                if qtype == "*":
                    query_all_types = True
                    #  since we are querying all types; create a set of domain_list to filter out double entries
                    domain_list = list(set(domain_list))
                else:
                    try:
                        qtype = getattr(q_types, qtype)
                    except AttributeError:
                        self.logger.warning(
                            f"Requesting a dns type ({qtype}) which is not in the q_types tuple; passing requested int"
                        )
                        pass

            if qtype == q_types.PTR and not query_all_types:
                domain_list = [self._convert_to_reverse_format(x) for x in domain_list]

            for domain in domain_list:
                if query_all_types:
                    if len(domain.split(".")) == 2:
                        q_type_list = [
                            x
                            for x in q_types.__dir__()
                            if not x.startswith("_")
                            and x not in ["index", "count"]
                            and x != "PTR"
                        ]
                    else:
                        q_type_list = ["A", "AAAA", "CNAME", "TXT"]

                    for q_type in q_type_list:
                        resources.append(
                            f"dns-query?name={domain}&type={getattr(q_types, q_type)}"
                        )
                else:
                    resources.append(f"dns-query?name={domain}&type={qtype}")
        data = await self.a_call(self.methods.GET, resources=resources)
        return data

    async def get_reverse_lookups(self, ips: List[str]) -> List[Response]:
        """
        Get the list of reverse DNS lookups.

        Args:
            ips: list of IP addresses to lookup

        Returns:
            List of `Response` objects
        """
        resources = [
            f"dns-query?name={self._convert_to_reverse_format(each)}&type=PTR"
            for each in ips
        ]
        data = await self.a_call(self.methods.GET, resources=resources)
        return data



from typing import Type, TypeVar

from dataclasses_json import DataClassJsonMixin

from nldcsc.http_apis.base_class.api_base_class import ApiBaseClass

from .objects import (
    NCSCFeedIndex,
    NCSCFeedInfo,
    NCSCFeedItems,
    NCSCFeedUpdateSequenceList,
    NCSCFeedUpdateTS,
    FBFErrors,
    FBFInfo,
)

P = TypeVar("P", bound=DataClassJsonMixin)


class FBFClient(ApiBaseClass):
    def __init__(
        self,
        baseurl: str,
        auth: str,
        api_path: str = None,
        proxies=None,
        user_agent="nldcsc",
        **kwargs,
    ):
        super().__init__(baseurl, api_path, proxies, user_agent, **kwargs)

        self.set_header_field("access-token", auth)

    def call(
        self,
        method=None,
        resource=None,
        data=None,
        timeout=60,
        return_response_object=False,
        stream=False,
        ignore_api_path=False,
        files=None,
        response_dataclass: Type[P] = None,
    ):
        data = super().call(
            method,
            resource,
            data,
            timeout,
            return_response_object,
            stream,
            ignore_api_path,
            files,
        )

        if response_dataclass:
            try:
                return response_dataclass.from_dict(data)
            except Exception:
                return FBFErrors.from_dict(data)

        return data

    def get_info(self):
        return self.call(self.methods.GET, response_dataclass=FBFInfo)

    def get_ncsc_feeds(self):
        resource = "ncsc/feeds"

        return self.call(
            self.methods.GET, resource=resource, response_dataclass=NCSCFeedInfo
        )

    def get_ncsc_feed_updates(self):
        resource = "ncsc/feeds/updates"

        return self.call(
            self.methods.GET,
            resource=resource,
            response_dataclass=NCSCFeedUpdateSequenceList,
        )

    def get_ncsc_feed_index(
        self,
        feed: str,
        client: str,
        start: int = None,
        stop: int = None,
        inclusive: bool = True,
    ):
        """
        Create a index from a ncsc feed

        Args:
            feed (str): name of feed
            client (str): client identifier
            start (int, optional): start timestamp of the first item. Defaults to None.
            stop (int, optional): timestamp of the last item. Defaults to None.
            inclusive (bool, optional): include the start time. Defaults to True.
        """
        resource = f"ncsc/feeds/request/{feed}/{client}"

        params = {}

        if start is not None:
            params["start_time"] = start

        if stop is not None:
            params["stop_time"] = stop

        if inclusive is not None:
            params["include_start_time"] = inclusive

        return self.call(
            self.methods.GET,
            resource=resource,
            data=params,
            response_dataclass=NCSCFeedIndex,
        )

    def get_ncsc_feed_index_items(self, index: str, limit: int = 1000, offset: int = 0):
        """
        Get items from a ncsc feed index

        Args:
            index (str): name of index
            limit (int, optional): amount of items. Defaults to 1000.
            offset (int, optional): item offset. Defaults to 0.
        """
        resource = f"ncsc/feeds//process/{index}"

        params = {"limit": limit, "offset": offset}

        return self.call(
            self.methods.GET,
            resource=resource,
            data=params,
            response_dataclass=NCSCFeedItems,
        )

    def iter_ncsc_feed_index_items(
        self, index: str, batch_size: int = 1000, offset: int = 0
    ):
        """
        iter through a ncsc index

        Args:
            index (str): name of index
            batch_size (int, optional): size to request per request. Defaults to 1000.
            offset (int, optional): starting offset. Defaults to 0.
        """
        batch = 0

        while True:
            batch = self.get_ncsc_feed_index_items(
                index, batch_size, offset + batch_size * batch
            )

            yield from batch.entries

            if batch.last:
                break

    def get_last_stop_ts(self, feed: str, client: str):
        resource = f"ncsc/feeds/last_update_ts/{feed}/{client}"

        return self.call(
            self.methods.GET, resource=resource, response_dataclass=NCSCFeedUpdateTS
        )



from dataclasses import dataclass, field
from inspect import signature
from typing import Optional

from dataclasses_json import config as json_config
from dataclasses_json import dataclass_json

from nldcsc.generic.utils import exclude_optional_dict
from nldcsc.plugins.vuln_tracker.objects.common import CVSS40, CVSS30, CVSS31, CVSS2
from nldcsc.plugins.vuln_tracker.objects.data_class_validations import Validations
from nldcsc.plugins.vuln_tracker.objects.osv import Affected


@dataclass_json
@dataclass
class Metadata(Validations):
    cveId: str
    assignerOrgId: str
    state: str

    assignerShortName: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    requesterUserId: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    dateUpdated: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    serial: Optional[int] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    dateReserved: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    datePublished: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    dateRejected: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )


@dataclass_json
@dataclass
class CNAInfo(Validations):
    orgId: str

    shortName: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    dateUpdated: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )


@dataclass_json
@dataclass
class Media(Validations):
    type: str
    value: str

    base64: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )


@dataclass_json
@dataclass
class Description(Validations):
    lang: str
    value: str

    supportingMedia: Optional[Media] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if hasattr(self, "supportingMedia"):
            if isinstance(self.supportingMedia, dict):
                self.supportingMedia = Media(**self.supportingMedia)


@dataclass_json
@dataclass
class ProgramRoutine(Validations):
    name: str


@dataclass_json
@dataclass
class Change(Validations):
    at: str
    status: str


@dataclass_json
@dataclass
class Version(Validations):
    version: str
    status: str

    versionType: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    lessThan: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    lessThanOrEqual: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    changes: Optional[list[Change]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if isinstance(self.changes, list):
            self.changes = [Change(**dict(x)) for x in self.changes]


@dataclass_json
@dataclass
class Affected(Validations):
    vendor: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    product: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    collectionURL: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    packageName: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    cpes: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    modules: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    programFiles: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    programRoutines: Optional[list[ProgramRoutine]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    platforms: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    repo: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    defaultStatus: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    versions: Optional[list[Version]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if isinstance(self.programRoutines, list):
            self.programRoutines = [
                ProgramRoutine(**dict(x)) for x in self.programRoutines
            ]

        if isinstance(self.versions, list):
            self.versions = [Version(**dict(x)) for x in self.versions]


@dataclass_json
@dataclass
class Reference(Validations):
    url: str

    name: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    tags: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )


@dataclass_json
@dataclass
class ProblemTypeDescription(Validations):
    lang: str
    description: str

    cweId: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    type: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    references: Optional[Reference] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if hasattr(self, "references"):
            if isinstance(self.references, dict):
                self.references = Reference(**self.references)


@dataclass_json
@dataclass
class ProblemType(Validations):
    descriptions: list[ProblemTypeDescription]

    def __post_init__(self):
        if isinstance(self.descriptions, list):
            self.descriptions = [
                ProblemTypeDescription(**dict(x)) for x in self.descriptions
            ]


@dataclass_json
@dataclass
class Impact(Validations):
    descriptions: list[Description]

    capecId: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if isinstance(self.descriptions, list):
            self.descriptions = [Description(**dict(x)) for x in self.descriptions]


@dataclass_json
@dataclass
class Scenario(Validations):
    lang: str
    value: str


@dataclass_json
@dataclass
class Other(Validations):
    type: str
    content: dict


@dataclass_json
@dataclass
class Metric(Validations):
    format: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    scenarios: Optional[list[Scenario]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    cvssV4_0: Optional[CVSS40] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    cvssV3_1: Optional[CVSS31] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    cvssV3_0: Optional[CVSS30] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    cvssV2_0: Optional[CVSS2] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    other: Optional[Other] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if hasattr(self, "scenarios"):
            if isinstance(self.scenarios, list):
                self.scenarios = [Scenario(**dict(x)) for x in self.scenarios]

        if hasattr(self, "cvssV4_0"):
            if isinstance(self.cvssV4_0, dict):
                self.cvssV4_0 = CVSS40(**self.cvssV4_0)

        if hasattr(self, "cvssV3_1"):
            if isinstance(self.cvssV3_1, dict):
                self.cvssV3_1 = CVSS31(**self.cvssV3_1)

        if hasattr(self, "cvssV3_0"):
            if isinstance(self.cvssV3_0, dict):
                self.cvssV3_0 = CVSS30(**self.cvssV3_0)

        if hasattr(self, "cvssV2_0"):
            if isinstance(self.cvssV2_0, dict):
                self.cvssV2_0 = CVSS2(**self.cvssV2_0)

        if hasattr(self, "other"):
            if isinstance(self.other, dict):
                self.other = Other(**self.other)


@dataclass_json
@dataclass
class Timeline(Validations):
    time: str
    lang: str
    value: str


@dataclass_json
@dataclass
class Credit(Validations):
    lang: str
    value: str

    user: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    type: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )


@dataclass_json
@dataclass
class TaxonomyRelation(Validations):
    taxonomyId: str
    relationshipId: str
    relationshipValue: str


@dataclass_json
@dataclass
class TaxonomyMapping(Validations):
    taxonomyName: str
    taxonomyRelations: list[TaxonomyRelation]

    taxonomyVersion: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if isinstance(self.taxonomyRelations, list):
            self.taxonomyRelations = [
                TaxonomyRelation(**dict(x)) for x in self.taxonomyRelations
            ]


@dataclass_json
@dataclass
class CNARejected(Validations):
    providerMetadata: CNAInfo
    rejectedReasons: list[Description] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    replacedBy: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    @classmethod
    def from_kwargs(cls, **kwargs):
        # fetch the constructor's signature
        cls_fields = {field for field in signature(cls).parameters}

        # split the kwargs into native ones and new ones
        native_args, new_args = {}, {}
        for name, val in kwargs.items():
            if name in cls_fields:
                native_args[name] = val
            else:
                new_args[name] = val

        # use the native ones to create the class ...
        ret = cls(**native_args)

        # ... and add the new ones by hand
        for new_name, new_val in new_args.items():
            setattr(ret, new_name, new_val)
        return ret

    def to_rss_format(self):
        return {
            "rejectedReasons": f"{[x.value for x in self.rejectedReasons]}",
        }

    def __post_init__(self):
        if isinstance(self.providerMetadata, dict):
            self.providerMetadata = CNAInfo(**self.providerMetadata)

        if isinstance(self.rejectedReasons, list):
            self.rejectedReasons = [
                Description(**dict(x)) for x in self.rejectedReasons
            ]


@dataclass_json
@dataclass
class CNAPublished(Validations):
    providerMetadata: CNAInfo
    descriptions: list[Description]
    affected: list[Affected]
    references: list[Reference]

    dateAssigned: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    datePublic: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    title: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    problemTypes: Optional[list[ProblemType]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    impacts: Optional[list[Impact]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    metrics: Optional[list[Metric]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    configurations: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    workarounds: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    solutions: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    exploits: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    timeline: Optional[list[Timeline]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    credits: Optional[list[Credit]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    source: Optional[dict] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    tags: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    taxonomyMappings: Optional[list[TaxonomyMapping]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    @classmethod
    def from_kwargs(cls, **kwargs):
        # fetch the constructor's signature
        cls_fields = {field for field in signature(cls).parameters}

        # split the kwargs into native ones and new ones
        native_args, new_args = {}, {}
        for name, val in kwargs.items():
            if name in cls_fields:
                native_args[name] = val
            else:
                new_args[name] = val

        # use the native ones to create the class ...
        ret = cls(**native_args)

        # ... and add the new ones by hand
        for new_name, new_val in new_args.items():
            setattr(ret, new_name, new_val)
        return ret

    def to_rss_format(self):

        if len(self.affected) == 1:
            return {
                "product": self.affected[0].product,
                "vendor": self.affected[0].vendor,
                "versions": f"{[{'status': x.status, 'version': x.version} for x in self.affected[0].versions]}",
                "descriptions": f"{[x.value for x in self.descriptions]}",
                "references": f"{[x.url for x in self.references]}",
            }
        else:
            return {
                "affected_products": f"Multiple products / vendors: {len(self.affected)}",
                "descriptions": f"{[x.value for x in self.descriptions]}",
                "references": f"{[x.url for x in self.references]}",
            }

    def __post_init__(self):
        if isinstance(self.providerMetadata, dict):
            self.providerMetadata = CNAInfo(**self.providerMetadata)

        if isinstance(self.descriptions, list):
            self.descriptions = [Description(**dict(x)) for x in self.descriptions]

        if isinstance(self.affected, list):
            self.affected = [Affected(**dict(x)) for x in self.affected]

        if isinstance(self.references, list):
            self.references = [Reference(**dict(x)) for x in self.references]

        if hasattr(self, "problemTypes"):
            if isinstance(self.problemTypes, list):
                self.problemTypes = [ProblemType(**dict(x)) for x in self.problemTypes]

        if hasattr(self, "impacts"):
            if isinstance(self.impacts, list):
                self.impacts = [Impact(**dict(x)) for x in self.impacts]

        if hasattr(self, "metrics"):
            if isinstance(self.metrics, list):
                self.metrics = [Metric(**dict(x)) for x in self.metrics]

        if hasattr(self, "configurations"):
            if isinstance(self.configurations, list):
                self.configurations = [
                    Description(**dict(x)) for x in self.configurations
                ]

        if hasattr(self, "workarounds"):
            if isinstance(self.workarounds, list):
                self.workarounds = [Description(**dict(x)) for x in self.workarounds]

        if hasattr(self, "solutions"):
            if isinstance(self.solutions, list):
                self.solutions = [Description(**dict(x)) for x in self.solutions]

        if hasattr(self, "exploits"):
            if isinstance(self.exploits, list):
                self.exploits = [Description(**dict(x)) for x in self.exploits]

        if hasattr(self, "timeline"):
            if isinstance(self.timeline, list):
                self.timeline = [Timeline(**dict(x)) for x in self.timeline]

        if hasattr(self, "credits"):
            if isinstance(self.credits, list):
                self.credits = [Credit(**dict(x)) for x in self.credits]

        if hasattr(self, "taxonomyMappings"):
            if isinstance(self.taxonomyMappings, list):
                self.taxonomyMappings = [
                    TaxonomyMapping(**dict(x)) for x in self.taxonomyMappings
                ]


@dataclass_json
@dataclass
class ADP(Validations):
    providerMetadata: CNAInfo

    descriptions: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    affected: Optional[list[Affected]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    references: Optional[list[Reference]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    dateAssigned: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    datePublic: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    title: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    problemTypes: Optional[list[ProblemType]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    impacts: Optional[list[Impact]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    metrics: Optional[list[Metric]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    configurations: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    workarounds: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    solutions: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    exploits: Optional[list[Description]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    timeline: Optional[list[Timeline]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    credits: Optional[list[Credit]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    source: Optional[dict] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    tags: Optional[list[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    taxonomyMappings: Optional[list[TaxonomyMapping]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    # this one might need some dynamic handling, all object with starting x (Pattern property)
    x_generator: Optional[dict] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if isinstance(self.providerMetadata, dict):
            self.providerMetadata = CNAInfo(**self.providerMetadata)

        if hasattr(self, "descriptions"):
            if isinstance(self.descriptions, list):
                self.descriptions = [Description(**dict(x)) for x in self.descriptions]

        if hasattr(self, "affected"):
            if isinstance(self.affected, list):
                self.affected = [Affected(**dict(x)) for x in self.affected]

        if hasattr(self, "references"):
            if isinstance(self.references, list):
                self.references = [Reference(**dict(x)) for x in self.references]

        if hasattr(self, "problemTypes"):
            if isinstance(self.problemTypes, list):
                self.problemTypes = [ProblemType(**dict(x)) for x in self.problemTypes]

        if hasattr(self, "impacts"):
            if isinstance(self.impacts, list):
                self.impacts = [Impact(**dict(x)) for x in self.impacts]

        if hasattr(self, "metrics"):
            if isinstance(self.metrics, list):
                self.metrics = [Metric(**dict(x)) for x in self.metrics]

        if hasattr(self, "configurations"):
            if isinstance(self.configurations, list):
                self.configurations = [
                    Description(**dict(x)) for x in self.configurations
                ]

        if hasattr(self, "workarounds"):
            if isinstance(self.workarounds, list):
                self.workarounds = [Description(**dict(x)) for x in self.workarounds]

        if hasattr(self, "solutions"):
            if isinstance(self.solutions, list):
                self.solutions = [Description(**dict(x)) for x in self.solutions]

        if hasattr(self, "exploits"):
            if isinstance(self.exploits, list):
                self.exploits = [Description(**dict(x)) for x in self.exploits]

        if hasattr(self, "timeline"):
            if isinstance(self.timeline, list):
                self.timeline = [Timeline(**dict(x)) for x in self.timeline]

        if hasattr(self, "credits"):
            if isinstance(self.credits, list):
                self.credits = [Credit(**dict(x)) for x in self.credits]

        if hasattr(self, "taxonomyMappings"):
            if isinstance(self.taxonomyMappings, list):
                self.taxonomyMappings = [
                    TaxonomyMapping(**dict(x)) for x in self.taxonomyMappings
                ]


@dataclass_json
@dataclass
class Containers(Validations):
    cna: CNAPublished | CNARejected
    cna_type: str = "published"

    adp: Optional[ADP] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    def __post_init__(self):
        if isinstance(self.cna, dict):
            if self.cna_type == "published":
                self.cna = CNAPublished.from_kwargs(**self.cna)
            else:
                self.cna = CNARejected.from_kwargs(**self.cna)

        if hasattr(self, "adp"):
            if isinstance(self.adp, dict):
                self.adp = ADP(**self.adp)

        delattr(self, "cna_type")


@dataclass_json
@dataclass
class CveRecord(Validations):
    dataType: str
    dataVersion: str
    cveMetadata: Metadata
    containers: Containers

    @property
    def cve_id(self) -> str:
        return f"{self.cveMetadata.cveId}"

    @property
    def state(self) -> str:
        return f"{self.cveMetadata.state}"

    def to_rss_format(self) -> dict:

        ret_dict = {
            "cve_id": self.cve_id,
            "state": self.state,
            "shortName": self.containers.cna.providerMetadata.shortName,
        }

        ret_dict.update(self.containers.cna.to_rss_format())

        return ret_dict

    def __repr__(self) -> str:
        return f"CVE_RECORD({self.state}[{self.cve_id}])"

    def __post_init__(self):
        if isinstance(self.cveMetadata, dict):
            self.cveMetadata = Metadata(**self.cveMetadata)

        if isinstance(self.containers, dict):
            self.containers = Containers(
                **self.containers, cna_type=self.cveMetadata.state.lower()
            )



import re
from typing import Callable, Iterable

from flask import request
from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy.query import Query
from sqlalchemy import or_, and_
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import BinaryExpression

from nldcsc.datatables.server_side_dt import ServerSideDataTable
from nldcsc.generic.times import timestringTOtimestamp
from nldcsc.generic.utils import str2bool


class SQLServerSideDataTable(ServerSideDataTable):
    """
    This class holds all the logic for enabling and handling server side DataTables within the application

    """

    def __init__(
        self,
        request: request,
        backend: SQLAlchemy,
        target_model: str,
        model_mapping: dict,
        additional_filters: list[BinaryExpression] = None,
        use_column_filters: bool = False,
        custom_column_filters: dict[
            str,
            Callable[[InstrumentedAttribute, str, bool], Iterable[BinaryExpression]],
        ] = None,
        query_hook: Callable[[Query], Query] = None,
        **kwargs,
    ):
        """
        Sql server side datatable.

        Generates a response needed from a datatable request.

        Args:
            request (request): The request the datatable info is in.
            backend (SQLAlchemy): The backend to use to retrieve data from.
            target_model (str): The targeted model to fetch.
            model_mapping (dict): Mapping of models to their class.
            additional_filters (list[BinaryExpression], optional): list of additional filter to apply. Defaults to None.
            use_column_filters (bool, optional): apply filters on the column level. Defaults to False.
            custom_column_filters (dict[ str, Callable[[InstrumentedAttribute, str, bool], BinaryExpression] ], optional): apply a custom filter to specific columns. Defaults to None.
            query_hook (Callable[[Query], Query], optional): hook that is called before the query is submitted, gives the ability to modify the query. Defaults to None.
        """
        if additional_filters is None:
            additional_filters = []

        if custom_column_filters is None:
            custom_column_filters = {}

        self.target_model = target_model
        self.additional_filters = additional_filters
        self.use_column_filters = use_column_filters
        self.custom_column_filters = custom_column_filters
        self.query_hook = query_hook

        self.models = model_mapping

        super().__init__(request, backend, **kwargs)

    def post_fetch_processing(self):
        pass

    def get_results_and_meta(self):
        if len(self.sort) == 0:
            self.sort = ""
        else:
            self.sort = self.create_sort_list()

        total_query = self.models[self.target_model].query
        for query_filter in self.additional_filters:
            total_query = total_query.filter(query_filter)
        self.total = total_query.count()

        self.filtered = []

        if self.use_column_filters:
            column_filters = self.column_data_filter()
            global_filters = self.data_filter()
            if len(column_filters) != 0 and len(global_filters) != 0:
                self.filtered.append(and_(*column_filters, or_(*global_filters)))
            elif len(column_filters) != 0:
                self.filtered.append(and_(*column_filters))
            elif len(global_filters) != 0:
                self.filtered.append(or_(*global_filters))
        else:
            global_filters = self.data_filter()
            if len(global_filters) != 0:
                self.filtered.append(or_(*global_filters))

        self.fetch_results()

        if len(self.filtered) != 0:
            filter_query = self.models[self.target_model].query
            for query_filter in self.additional_filters:
                filter_query = filter_query.filter(query_filter)

            filter_query = filter_query.filter(*self.filtered)
            self.total_filtered = filter_query.count()
        else:
            self.total_filtered = self.total

    def fetch_results(self):
        """
        Method responsible for querying the backend and fetching the results from the database
        """
        query: Query = self.models[self.target_model].query

        for query_filter in self.additional_filters:
            query = query.filter(query_filter)

        if len(self.filtered) != 0:
            query = query.filter(*self.filtered)

        if self.query_hook:
            query = self.query_hook(query)

        data_objects = (
            query.order_by(*self.sort)
            .offset(self.data_length.start)
            .limit(self.data_length.length)
            .all()
        )

        self.results = [x.to_data_dict() for x in data_objects]

    def create_sort_list(self):
        ret_list = []

        for each in self.sort:
            c = getattr(self.models[self.target_model], each[0], None)

            if not c:
                continue

            if each[1] == "asc":
                ret_list.append(c.asc())
            elif each[1] == "desc":
                ret_list.append(c.desc())

        return ret_list

    @staticmethod
    def default_filter(
        col: InstrumentedAttribute,
        filter_val: str,
        is_regex: bool,
        check_date: bool = True,
    ) -> list[BinaryExpression]:
        def is_date(value: str) -> int | str:
            if not check_date:
                return value

            is_date = timestringTOtimestamp(value)

            if is_date is not False:
                return is_date

            return value

        filter_args = []

        if is_regex:
            try:
                negate_search = False

                if filter_val[0] == "~":
                    filter_val = filter_val[1:]
                    negate_search = True

                re.compile(filter_val)

                if negate_search:
                    filter_args.append(~col.regexp_match(filter_val))
                else:
                    filter_args.append(col.regexp_match(filter_val))

                return filter_args
            except re.error:
                if negate_search:
                    filter_val = "~" + filter_val

        if "<>" in filter_val:
            start, end = filter_val.split("<>", 1)

            start = is_date(start.strip())
            end = is_date(end.strip())

            filter_args.append(col > start)
            filter_args.append(col < end)
        elif filter_val[0] == ">":
            filter_val = is_date(filter_val[1:].strip())
            filter_args.append(col > filter_val)
        elif filter_val[0] == "<":
            filter_val = is_date(filter_val[1:].strip())
            filter_args.append(col < filter_val)
        elif filter_val[0] == "~":
            if len(filter_val.strip()) == 1:
                filter_args.append(col == None)  # noqa: E711
            else:
                filter_val = filter_val[1:].strip()
                filter_args.append(~col.like(f"%{filter_val}%"))
        else:
            filter_args.append(col.like(f"%{filter_val}%"))

        return filter_args

    def get_filter_args(
        self, column: str, filter_val: str, is_regex: bool, check_date: bool = True
    ) -> list:
        """
        parses search values to valid search expressions for sqlalchemy.

        when searching without regex the <>, >, < and ~ operators are supported.
        where <> searches between two values, > searches for larger values, < searches for smaller values and ~
        reverses the search operation. If only ~ is searched it is treated as a NULL filter, meaning column must
        be holding value NULL.

        when searching with regex only ~ is supported, reversing the search to a negate search.

        If a custom filter is given for a column that function will be called with the column, filter_val and is_regex.
        If the custom filter can't filter the values it can raise a ValueError to proceed with the normal filters.
        The custom filter should return a list of filters.
            Example:
                def my_custom_filter(col: InstrumentedAttribute, filter_val: str, is_regex: bool):
                    if filter_val.startswith("A"):
                        return [col != "A", col != "a"]
                    else:
                        raise ValueError

        Args:
            column (str): str value representing the ORM column to filter on.
            filter_val (str): value to search on.
            is_regex (bool): if the search is a regex search or a regular search
            check_date (bool, optional): will check if strings represent a date and convert it to a timestamp value.
            Defaults to True.

        Returns:
            list: List of filter values
        """
        filter_val = str(filter_val)

        custom_filter = self.custom_column_filters.get(column, None)

        filter_args = []

        try:
            col = getattr(self.models[self.target_model], column)
        except AttributeError:
            return filter_args

        if custom_filter is not None:
            try:
                # noinspection PyCallingNonCallable
                filter_args.extend(custom_filter(col, filter_val, is_regex))
                return filter_args
            except ValueError:
                # If the custom filter throws a ValueError we apply the normal filters
                pass

        filter_args.extend(self.default_filter(col, filter_val, is_regex, check_date))

        return filter_args

    def data_filter(self) -> list:
        """
        method responsible for retrieving filter values based on the table wide search value for every supported column.

        Returns:
            list: list containing prepared filters.
        """

        search_val = self.request_values["search[value]"]
        is_regex = str2bool(
            self.request_values.get(key="search[regex]", default="false")
        )

        search_args = []

        if search_val != "" and search_val != "$" and search_val != "~":
            for col in self.columns:
                if str2bool(self.columns[col]["searchable"]):
                    search_args.extend(
                        self.get_filter_args(
                            self.columns[col]["data"], search_val, is_regex
                        )
                    )

        return search_args

    def column_data_filter(self) -> list:
        """
        method responsible for retrieving filter values based on a per column based search value.

        Returns:
            list: list containing prepared filters.
        """
        search_args = []

        for col in self.columns:
            search_val_dict = self.columns[col].get(
                "search", {"regex": "false", "value": ""}
            )

            if search_val_dict["value"] in ["", "$"]:
                continue

            search_val_dict["regex"] = str2bool(search_val_dict["regex"])

            search_args.extend(
                self.get_filter_args(
                    self.columns[col]["data"],
                    search_val_dict["value"],
                    search_val_dict["regex"],
                )
            )

        return search_args
